@@ -175,11 +175,14 @@ void InputSplitBase::InitInputFileInfo(const std::string& uri,
 }
 
 size_t InputSplitBase::Read(void *ptr, size_t size) {
+  const bool is_text_parser = this->IsTextParser();
+
+  if (fs_ == NULL) {
+    return 0;
+  }
   if (offset_begin_ >= offset_end_) return 0;
   if (offset_curr_ +  size > offset_end_) {
-    // allow for extra newlines between files
-    const size_t num_extra_eol = files_.size() - file_ptr_ - 1;
-    size = offset_end_ - offset_curr_ + num_extra_eol;
+    size = offset_end_ - offset_curr_;
   }
   if (size == 0) return 0;
   size_t nleft = size;
@@ -190,7 +193,11 @@ size_t InputSplitBase::Read(void *ptr, size_t size) {
     offset_curr_ += n;
     if (nleft == 0) break;
     if (n == 0) {
-      buf[0] = '\n'; ++buf; --nleft;  // insert a newline between files
+      if (is_text_parser) {
+        // Insert a newline between files to handle files with NOEOL.
+        // See https://github.com/dmlc/dmlc-core/pull/385 for explanation.
+        buf[0] = '\n'; ++buf; --nleft;
+      }
       if (offset_curr_ != file_offset_[file_ptr_ + 1]) {
         LOG(ERROR) << "curr=" << offset_curr_
                    << ",begin=" << offset_begin_
@@ -225,20 +232,29 @@ bool InputSplitBase::ReadChunk(void *buf, size_t *size) {
                             max_size - olen);
   nread += olen;
   if (nread == 0) return false;
-  if (nread != max_size) {
-    *size = nread;
-    return true;
+  if (this->IsTextParser()) {
+    if (nread == olen) {
+      // Insert a newline between files to handle files with NOEOL.
+      // See https://github.com/dmlc/dmlc-core/pull/452 for explanation.
+      char *bufptr = reinterpret_cast<char*>(buf);
+      bufptr[nread] = '\n';
+      nread++;
+    }
   } else {
-    const char *bptr = reinterpret_cast<const char*>(buf);
-    // return the last position where a record starts
-    const char *bend = this->FindLastRecordBegin(bptr, bptr + max_size);
-    *size = bend - bptr;
-    overflow_.resize(max_size - *size);
-    if (overflow_.length() != 0) {
-      std::memcpy(BeginPtr(overflow_), bend, overflow_.length());
+    if (nread != max_size) {
+      *size = nread;
+      return true;
     }
-    return true;
   }
+  const char *bptr = reinterpret_cast<const char*>(buf);
+  // return the last position where a record starts
+  const char *bend = this->FindLastRecordBegin(bptr, bptr + nread);
+  *size = bend - bptr;
+  overflow_.resize(nread - *size);
+  if (overflow_.length() != 0) {
+    std::memcpy(BeginPtr(overflow_), bend, overflow_.length());
+  }
+  return true;
 }
 
 bool InputSplitBase::Chunk::Load(InputSplitBase *split, size_t buffer_size) {
