@@ -71,7 +71,7 @@ class ColumnMatrix {
   // construct column matrix from GHistIndexMatrix
   inline void Init(const GHistIndexMatrix& gmat,
                    double  sparse_threshold) {
-    const int32_t nfeature = static_cast<int32_t>(gmat.cut.row_ptr.size() - 1);
+    const int32_t nfeature = static_cast<int32_t>(gmat.cut.Ptrs().size() - 1);
     const size_t nrow = gmat.row_ptr.size() - 1;
 
     // identify type of each column
@@ -80,8 +80,8 @@ class ColumnMatrix {
     std::fill(feature_counts_.begin(), feature_counts_.end(), 0);
 
     uint32_t max_val = std::numeric_limits<uint32_t>::max();
-    for (bst_uint fid = 0; fid < nfeature; ++fid) {
-      CHECK_LE(gmat.cut.row_ptr[fid + 1] - gmat.cut.row_ptr[fid], max_val);
+    for (int32_t fid = 0; fid < nfeature; ++fid) {
+      CHECK_LE(gmat.cut.Ptrs()[fid + 1] - gmat.cut.Ptrs()[fid], max_val);
     }
 
     gmat.GetFeatureCounts(&feature_counts_[0]);
@@ -119,8 +119,8 @@ class ColumnMatrix {
 
     // store least bin id for each feature
     index_base_.resize(nfeature);
-    for (bst_uint fid = 0; fid < nfeature; ++fid) {
-      index_base_[fid] = gmat.cut.row_ptr[fid];
+    for (int32_t fid = 0; fid < nfeature; ++fid) {
+      index_base_[fid] = gmat.cut.Ptrs()[fid];
     }
 
     // pre-fill index_ for dense columns
@@ -147,9 +147,9 @@ class ColumnMatrix {
       size_t fid = 0;
       for (size_t i = ibegin; i < iend; ++i) {
         const uint32_t bin_id = gmat.index[i];
-        while (bin_id >= gmat.cut.row_ptr[fid + 1]) {
-          ++fid;
-        }
+        auto iter = std::upper_bound(gmat.cut.Ptrs().cbegin() + fid,
+                                     gmat.cut.Ptrs().cend(), bin_id);
+        fid = std::distance(gmat.cut.Ptrs().cbegin(), iter) - 1;
         if (type_[fid] == kDenseColumn) {
           uint32_t* begin = &index_[boundary_[fid].index_begin];
           begin[rid] = bin_id - index_base_[fid];
