@@ -6,7 +6,9 @@
 #include <xgboost/objective.h>
 #include <dmlc/registry.h>
 
-#include "../common/host_device_vector.h"
+#include <sstream>
+
+#include "xgboost/host_device_vector.h"
 
 namespace dmlc {
 DMLC_REGISTRY_ENABLE(::xgboost::ObjFunctionReg);
@@ -14,15 +16,19 @@ DMLC_REGISTRY_ENABLE(::xgboost::ObjFunctionReg);
 
 namespace xgboost {
 // implement factory functions
-ObjFunction* ObjFunction::Create(const std::string& name) {
+ObjFunction* ObjFunction::Create(const std::string& name, GenericParameter const* tparam) {
   auto *e = ::dmlc::Registry< ::xgboost::ObjFunctionReg>::Get()->Find(name);
   if (e == nullptr) {
+    std::stringstream ss;
     for (const auto& entry : ::dmlc::Registry< ::xgboost::ObjFunctionReg>::List()) {
-      LOG(INFO) << "Objective candidate: " << entry->name;
+      ss << "Objective candidate: " << entry->name << "\n";
     }
-    LOG(FATAL) << "Unknown objective function " << name;
+    LOG(FATAL) << "Unknown objective function: `" << name << "`\n"
+               << ss.str();
   }
-  return (e->body)();
+  auto pobj = (e->body)();
+  pobj->tparam_ = tparam;
+  return pobj;
 }
 
 }  // namespace xgboost
@@ -34,11 +40,12 @@ namespace obj {
 DMLC_REGISTRY_LINK_TAG(regression_obj_gpu);
 DMLC_REGISTRY_LINK_TAG(hinge_obj_gpu);
 DMLC_REGISTRY_LINK_TAG(multiclass_obj_gpu);
+DMLC_REGISTRY_LINK_TAG(rank_obj_gpu);
 #else
 DMLC_REGISTRY_LINK_TAG(regression_obj);
 DMLC_REGISTRY_LINK_TAG(hinge_obj);
 DMLC_REGISTRY_LINK_TAG(multiclass_obj);
-#endif  // XGBOOST_USE_CUDA
 DMLC_REGISTRY_LINK_TAG(rank_obj);
+#endif  // XGBOOST_USE_CUDA
 }  // namespace obj
 }  // namespace xgboost
