@@ -1,17 +1,21 @@
 /*!
- * Copyright 2014 by Contributors
+ * Copyright 2014-2019 by Contributors
  * \file updater_colmaker.cc
  * \brief use columnwise update to construct a tree
  * \author Tianqi Chen
  */
 #include <rabit/rabit.h>
-#include <xgboost/tree_updater.h>
 #include <memory>
 #include <vector>
 #include <cmath>
 #include <algorithm>
 
-#include "./param.h"
+#include "xgboost/parameter.h"
+#include "xgboost/tree_updater.h"
+#include "xgboost/logging.h"
+#include "xgboost/json.h"
+#include "param.h"
+#include "constraints.h"
 #include "../common/random.h"
 #include "../common/bitmap.h"
 #include "split_evaluator.h"
@@ -21,26 +25,76 @@ namespace tree {
 
 DMLC_REGISTRY_FILE_TAG(updater_colmaker);
 
+struct ColMakerTrainParam : XGBoostParameter<ColMakerTrainParam> {
+  // speed optimization for dense column
+  float opt_dense_col;
+  DMLC_DECLARE_PARAMETER(ColMakerTrainParam) {
+    DMLC_DECLARE_FIELD(opt_dense_col)
+        .set_range(0.0f, 1.0f)
+        .set_default(1.0f)
+        .describe("EXP Param: speed optimization for dense column.");
+  }
+
+  /*! \brief whether need forward small to big search: default right */
+  inline bool NeedForwardSearch(int default_direction, float col_density,
+                                bool indicator) const {
+    return default_direction == 2 ||
+           (default_direction == 0 && (col_density < opt_dense_col) &&
+            !indicator);
+  }
+  /*! \brief whether need backward big to small search: default left */
+  inline bool NeedBackwardSearch(int default_direction) const {
+    return default_direction != 2;
+  }
+};
+
+DMLC_REGISTER_PARAMETER(ColMakerTrainParam);
+
 /*! \brief column-wise update to construct a tree */
 class ColMaker: public TreeUpdater {
  public:
-  void Init(const std::vector<std::pair<std::string, std::string> >& args) override {
-    param_.InitAllowUnknown(args);
-    spliteval_.reset(SplitEvaluator::Create(param_.split_evaluator));
-    spliteval_->Init(args);
+  void Configure(const Args& args) override {
+    param_.UpdateAllowUnknown(args);
+    colmaker_param_.UpdateAllowUnknown(args);
+    if (!spliteval_) {
+      spliteval_.reset(SplitEvaluator::Create(param_.split_evaluator));
+    }
+    spliteval_->Init(&param_);
+  }
+
+  void LoadConfig(Json const& in) override {
+    auto const& config = get<Object const>(in);
+    fromJson(config.at("train_param"), &this->param_);
+    fromJson(config.at("colmaker_train_param"), &this->colmaker_param_);
+  }
+  void SaveConfig(Json* p_out) const override {
+    auto& out = *p_out;
+    out["train_param"] = toJson(param_);
+    out["colmaker_train_param"] = toJson(colmaker_param_);
+  }
+
+  char const* Name() const override {
+    return "grow_colmaker";
   }
 
   void Update(HostDeviceVector<GradientPair> *gpair,
               DMatrix* dmat,
               const std::vector<RegTree*> &trees) override {
+    if (rabit::IsDistributed()) {
+      LOG(FATAL) << "Updater `grow_colmaker` or `exact` tree method doesn't "
+                    "support distributed training.";
+    }
     // rescale learning rate according to size of trees
     float lr = param_.learning_rate;
     param_.learning_rate = lr / trees.size();
+    interaction_constraints_.Configure(param_, dmat->Info().num_row_);
     // build tree
     for (auto tree : trees) {
       Builder builder(
         param_,
-        std::unique_ptr<SplitEvaluator>(spliteval_->GetHostClone()));
+        colmaker_param_,
+        std::unique_ptr<SplitEvaluator>(spliteval_->GetHostClone()),
+        interaction_constraints_);
       builder.Update(gpair->ConstHostVector(), dmat, tree);
     }
     param_.learning_rate = lr;
@@ -49,23 +103,22 @@ class ColMaker: public TreeUpdater {
  protected:
   // training parameter
   TrainParam param_;
+  ColMakerTrainParam colmaker_param_;
   // SplitEvaluator that will be cloned for each Builder
   std::unique_ptr<SplitEvaluator> spliteval_;
+
+  FeatureInteractionConstraintHost interaction_constraints_;
   // data structure
   /*! \brief per thread x per node entry to store tmp data */
   struct ThreadEntry {
     /*! \brief statistics of data */
     GradStats stats;
-    /*! \brief extra statistics of data */
-    GradStats stats_extra;
     /*! \brief last feature value scanned */
     bst_float last_fvalue;
-    /*! \brief first feature value scanned */
-    bst_float first_fvalue;
     /*! \brief current best solution */
     SplitEntry best;
     // constructor
-    ThreadEntry() : last_fvalue{0}, first_fvalue{0} {}
+    ThreadEntry() : last_fvalue{0} {}
   };
   struct NodeEntry {
     /*! \brief statics for node entry */
@@ -84,9 +137,13 @@ class ColMaker: public TreeUpdater {
    public:
     // constructor
     explicit Builder(const TrainParam& param,
-                     std::unique_ptr<SplitEvaluator> spliteval)
-        : param_(param), nthread_(omp_get_max_threads()),
-          spliteval_(std::move(spliteval)) {}
+                     const ColMakerTrainParam& colmaker_train_param,
+                     std::unique_ptr<SplitEvaluator> spliteval,
+                     FeatureInteractionConstraintHost _interaction_constraints)
+        : param_(param), colmaker_train_param_{colmaker_train_param},
+          nthread_(omp_get_max_threads()),
+          spliteval_(std::move(spliteval)),
+          interaction_constraints_{std::move(_interaction_constraints)} {}
     // update one tree, growing
     virtual void Update(const std::vector<GradientPair>& gpair,
                         DMatrix* p_fmat,
@@ -111,6 +168,7 @@ class ColMaker: public TreeUpdater {
                                snode_[nid].best.SplitIndex(),
                                snode_[cleft].weight,
                                snode_[cright].weight);
+          interaction_constraints_.Split(nid, snode_[nid].best.SplitIndex(), cleft, cright);
         }
         qexpand_ = newnodes;
         // if nothing left to be expand, break
@@ -133,21 +191,11 @@ class ColMaker: public TreeUpdater {
     inline void InitData(const std::vector<GradientPair>& gpair,
                          const DMatrix& fmat,
                          const RegTree& tree) {
-      CHECK_EQ(tree.param.num_nodes, tree.param.num_roots)
-          << "ColMaker: can only grow new tree";
-      const std::vector<unsigned>& root_index = fmat.Info().root_index_;
       {
         // setup position
         position_.resize(gpair.size());
         CHECK_EQ(fmat.Info().num_row_, position_.size());
-        if (root_index.size() == 0) {
-          std::fill(position_.begin(), position_.end(), 0);
-        } else {
-          for (size_t ridx = 0; ridx <  position_.size(); ++ridx) {
-            position_[ridx] = root_index[ridx];
-            CHECK_LT(root_index[ridx], (unsigned)tree.param.num_roots);
-          }
-        }
+        std::fill(position_.begin(), position_.end(), 0);
         // mark delete for the deleted datas
         for (size_t ridx = 0; ridx < position_.size(); ++ridx) {
           if (gpair[ridx].GetHess() < 0.0f) position_[ridx] = ~position_[ridx];
@@ -179,9 +227,7 @@ class ColMaker: public TreeUpdater {
       {
         // expand query
         qexpand_.reserve(256); qexpand_.clear();
-        for (int i = 0; i < tree.param.num_roots; ++i) {
-          qexpand_.push_back(i);
-        }
+        qexpand_.push_back(0);
       }
     }
     /*!
@@ -238,155 +284,11 @@ class ColMaker: public TreeUpdater {
         }
       }
     }
-    // parallel find the best split of current fid
-    // this function does not support nested functions
-    inline void ParallelFindSplit(const SparsePage::Inst &col,
-                                  bst_uint fid,
-                                 DMatrix *p_fmat,
-                                  const std::vector<GradientPair> &gpair) {
-      // TODO(tqchen): double check stats order.
-      const bool ind = col.size() != 0 && col[0].fvalue == col[col.size() - 1].fvalue;
-      bool need_forward = param_.NeedForwardSearch(p_fmat->GetColDensity(fid), ind);
-      bool need_backward = param_.NeedBackwardSearch(p_fmat->GetColDensity(fid), ind);
-      const std::vector<int> &qexpand = qexpand_;
-      #pragma omp parallel
-      {
-        const int tid = omp_get_thread_num();
-        std::vector<ThreadEntry> &temp = stemp_[tid];
-        // cleanup temp statistics
-        for (int j : qexpand) {
-          temp[j].stats = GradStats();
-        }
-        bst_uint step = (col.size() + this->nthread_ - 1) / this->nthread_;
-        bst_uint end = std::min(static_cast<bst_uint>(col.size()), step * (tid + 1));
-        for (bst_uint i = tid * step; i < end; ++i) {
-          const bst_uint ridx = col[i].index;
-          const int nid = position_[ridx];
-          if (nid < 0) continue;
-          const bst_float fvalue = col[i].fvalue;
-          if (temp[nid].stats.Empty()) {
-            temp[nid].first_fvalue = fvalue;
-          }
-          temp[nid].stats.Add(gpair[ridx]);
-          temp[nid].last_fvalue = fvalue;
-        }
-      }
-      // start collecting the partial sum statistics
-      auto nnode = static_cast<bst_omp_uint>(qexpand.size());
-      #pragma omp parallel for schedule(static)
-      for (bst_omp_uint j = 0; j < nnode; ++j) {
-        const int nid = qexpand[j];
-        GradStats sum, tmp, c;
-        for (int tid = 0; tid < this->nthread_; ++tid) {
-          tmp = stemp_[tid][nid].stats;
-          stemp_[tid][nid].stats = sum;
-          sum.Add(tmp);
-          if (tid != 0) {
-            std::swap(stemp_[tid - 1][nid].last_fvalue, stemp_[tid][nid].first_fvalue);
-          }
-        }
-        for (int tid = 0; tid < this->nthread_; ++tid) {
-          stemp_[tid][nid].stats_extra = sum;
-          ThreadEntry &e = stemp_[tid][nid];
-          bst_float fsplit;
-          if (tid != 0) {
-            if (stemp_[tid - 1][nid].last_fvalue != e.first_fvalue) {
-              fsplit = (stemp_[tid - 1][nid].last_fvalue + e.first_fvalue) * 0.5f;
-            } else {
-              continue;
-            }
-          } else {
-            fsplit = e.first_fvalue - kRtEps;
-          }
-          if (need_forward && tid != 0) {
-            c.SetSubstract(snode_[nid].stats, e.stats);
-            if (c.sum_hess >= param_.min_child_weight &&
-                e.stats.sum_hess >= param_.min_child_weight) {
-              auto loss_chg = static_cast<bst_float>(
-                  spliteval_->ComputeSplitScore(nid, fid, e.stats, c) -
-                  snode_[nid].root_gain);
-              e.best.Update(loss_chg, fid, fsplit, false, e.stats, c);
-            }
-          }
-          if (need_backward) {
-            tmp.SetSubstract(sum, e.stats);
-            c.SetSubstract(snode_[nid].stats, tmp);
-            if (c.sum_hess >= param_.min_child_weight &&
-                tmp.sum_hess >= param_.min_child_weight) {
-              auto loss_chg = static_cast<bst_float>(
-                  spliteval_->ComputeSplitScore(nid, fid, tmp, c) -
-                  snode_[nid].root_gain);
-              e.best.Update(loss_chg, fid, fsplit, true, tmp, c);
-            }
-          }
-        }
-        if (need_backward) {
-          tmp = sum;
-          ThreadEntry &e = stemp_[this->nthread_-1][nid];
-          c.SetSubstract(snode_[nid].stats, tmp);
-          if (c.sum_hess >= param_.min_child_weight &&
-              tmp.sum_hess >= param_.min_child_weight) {
-            auto loss_chg = static_cast<bst_float>(
-                spliteval_->ComputeSplitScore(nid, fid, tmp, c) -
-                snode_[nid].root_gain);
-            e.best.Update(loss_chg, fid, e.last_fvalue + kRtEps, true, tmp, c);
-          }
-        }
-      }
-      // rescan, generate candidate split
-      #pragma omp parallel
-      {
-        GradStats c, cright;
-        const int tid = omp_get_thread_num();
-        std::vector<ThreadEntry> &temp = stemp_[tid];
-        bst_uint step = (col.size() + this->nthread_ - 1) / this->nthread_;
-        bst_uint end = std::min(static_cast<bst_uint>(col.size()), step * (tid + 1));
-        for (bst_uint i = tid * step; i < end; ++i) {
-          const bst_uint ridx = col[i].index;
-          const int nid = position_[ridx];
-          if (nid < 0) continue;
-          const bst_float fvalue = col[i].fvalue;
-          // get the statistics of nid
-          ThreadEntry &e = temp[nid];
-          if (e.stats.Empty()) {
-            e.stats.Add(gpair[ridx]);
-            e.first_fvalue = fvalue;
-          } else {
-            // forward default right
-            if (fvalue != e.first_fvalue) {
-              if (need_forward) {
-                c.SetSubstract(snode_[nid].stats, e.stats);
-                if (c.sum_hess >= param_.min_child_weight &&
-                    e.stats.sum_hess >= param_.min_child_weight) {
-                  auto loss_chg = static_cast<bst_float>(
-                      spliteval_->ComputeSplitScore(nid, fid, e.stats, c) -
-                      snode_[nid].root_gain);
-                  e.best.Update(loss_chg, fid, (fvalue + e.first_fvalue) * 0.5f,
-                                false, e.stats, c);
-                }
-              }
-              if (need_backward) {
-                cright.SetSubstract(e.stats_extra, e.stats);
-                c.SetSubstract(snode_[nid].stats, cright);
-                if (c.sum_hess >= param_.min_child_weight &&
-                    cright.sum_hess >= param_.min_child_weight) {
-                  auto loss_chg = static_cast<bst_float>(
-                      spliteval_->ComputeSplitScore(nid, fid, c, cright) -
-                      snode_[nid].root_gain);
-                  e.best.Update(loss_chg, fid, (fvalue + e.first_fvalue) * 0.5f, true, c, cright);
-                }
-              }
-            }
-            e.stats.Add(gpair[ridx]);
-            e.first_fvalue = fvalue;
-          }
-        }
-      }
-    }
+
     // update enumeration solution
     inline void UpdateEnumeration(int nid, GradientPair gstats,
                                   bst_float fvalue, int d_step, bst_uint fid,
-                                  GradStats &c, std::vector<ThreadEntry> &temp) { // NOLINT(*)
+                                  GradStats &c, std::vector<ThreadEntry> &temp) const { // NOLINT(*)
       // get the statistics of nid
       ThreadEntry &e = temp[nid];
       // test if first hit, this is fine, because we set 0 during init
@@ -399,19 +301,31 @@ class ColMaker: public TreeUpdater {
             e.stats.sum_hess >= param_.min_child_weight) {
           c.SetSubstract(snode_[nid].stats, e.stats);
           if (c.sum_hess >= param_.min_child_weight) {
-            bst_float loss_chg;
+            bst_float loss_chg {0};
             if (d_step == -1) {
               loss_chg = static_cast<bst_float>(
                   spliteval_->ComputeSplitScore(nid, fid, c, e.stats) -
                   snode_[nid].root_gain);
-              e.best.Update(loss_chg, fid, (fvalue + e.last_fvalue) * 0.5f,
-                            d_step == -1, c, e.stats);
+              bst_float proposed_split = (fvalue + e.last_fvalue) * 0.5f;
+              if ( proposed_split == fvalue ) {
+                e.best.Update(loss_chg, fid, e.last_fvalue,
+                              d_step == -1, c, e.stats);
+              } else {
+                e.best.Update(loss_chg, fid, proposed_split,
+                              d_step == -1, c, e.stats);
+              }
             } else {
               loss_chg = static_cast<bst_float>(
                   spliteval_->ComputeSplitScore(nid, fid, e.stats, c) -
                   snode_[nid].root_gain);
-              e.best.Update(loss_chg, fid, (fvalue + e.last_fvalue) * 0.5f,
+              bst_float proposed_split = (fvalue + e.last_fvalue) * 0.5f;
+              if ( proposed_split == fvalue ) {
+                e.best.Update(loss_chg, fid, e.last_fvalue,
                             d_step == -1, e.stats, c);
+              } else {
+                e.best.Update(loss_chg, fid, proposed_split,
+                            d_step == -1, e.stats, c);
+              }
             }
           }
         }
@@ -421,12 +335,13 @@ class ColMaker: public TreeUpdater {
       }
     }
     // same as EnumerateSplit, with cacheline prefetch optimization
-    inline void EnumerateSplitCacheOpt(const Entry *begin,
-                                       const Entry *end,
-                                       int d_step,
-                                       bst_uint fid,
-                                       const std::vector<GradientPair> &gpair,
-                                       std::vector<ThreadEntry> &temp) { // NOLINT(*)
+    void EnumerateSplit(const Entry *begin,
+                        const Entry *end,
+                        int d_step,
+                        bst_uint fid,
+                        const std::vector<GradientPair> &gpair,
+                        std::vector<ThreadEntry> &temp)  const { // NOLINT(*)
+      CHECK(param_.cache_opt) << "Support for `cache_opt' is removed in 1.0.0";
       const std::vector<int> &qexpand = qexpand_;
       // clear all the temp statistics
       for (auto nid : qexpand) {
@@ -457,12 +372,13 @@ class ColMaker: public TreeUpdater {
         }
         for (i = 0, p = it; i < kBuffer; ++i, p += d_step) {
           const int nid = buf_position[i];
-          if (nid < 0) continue;
+          if (nid < 0 || !interaction_constraints_.Query(nid, fid)) { continue; }
           this->UpdateEnumeration(nid, buf_gpair[i],
                                   p->fvalue, d_step,
                                   fid, c, temp);
         }
       }
+
       // finish up the ending piece
       for (it = align_end, i = 0; it != end; ++i, it += d_step) {
         buf_position[i] = position_[it->index];
@@ -470,7 +386,7 @@ class ColMaker: public TreeUpdater {
       }
       for (it = align_end, i = 0; it != end; ++i, it += d_step) {
         const int nid = buf_position[i];
-        if (nid < 0) continue;
+        if (nid < 0 || !interaction_constraints_.Query(nid, fid)) { continue; }
         this->UpdateEnumeration(nid, buf_gpair[i],
                                 it->fvalue, d_step,
                                 fid, c, temp);
@@ -501,128 +417,43 @@ class ColMaker: public TreeUpdater {
       }
     }
 
-    // enumerate the split values of specific feature
-    inline void EnumerateSplit(const Entry *begin,
-                               const Entry *end,
-                               int d_step,
-                               bst_uint fid,
-                               const std::vector<GradientPair> &gpair,
-                               const MetaInfo &info,
-                               std::vector<ThreadEntry> &temp) { // NOLINT(*)
-      // use cacheline aware optimization
-      if (param_.cache_opt != 0) {
-        EnumerateSplitCacheOpt(begin, end, d_step, fid, gpair, temp);
-        return;
-      }
-      const std::vector<int> &qexpand = qexpand_;
-      // clear all the temp statistics
-      for (auto nid : qexpand) {
-        temp[nid].stats = GradStats();
-      }
-      // left statistics
-      GradStats c;
-      for (const Entry *it = begin; it != end; it += d_step) {
-        const bst_uint ridx = it->index;
-        const int nid = position_[ridx];
-        if (nid < 0) continue;
-        // start working
-        const bst_float fvalue = it->fvalue;
-        // get the statistics of nid
-        ThreadEntry &e = temp[nid];
-        // test if first hit, this is fine, because we set 0 during init
-        if (e.stats.Empty()) {
-          e.stats.Add(gpair[ridx]);
-          e.last_fvalue = fvalue;
-        } else {
-          // try to find a split
-          if (fvalue != e.last_fvalue &&
-              e.stats.sum_hess >= param_.min_child_weight) {
-            c.SetSubstract(snode_[nid].stats, e.stats);
-            if (c.sum_hess >= param_.min_child_weight) {
-              bst_float loss_chg;
-              if (d_step == -1) {
-                loss_chg = static_cast<bst_float>(
-                    spliteval_->ComputeSplitScore(nid, fid, c, e.stats) -
-                    snode_[nid].root_gain);
-                e.best.Update(loss_chg, fid, (fvalue + e.last_fvalue) * 0.5f,
-                              d_step == -1, c, e.stats);
-              } else {
-                loss_chg = static_cast<bst_float>(
-                    spliteval_->ComputeSplitScore(nid, fid, e.stats, c) -
-                    snode_[nid].root_gain);
-                e.best.Update(loss_chg, fid, (fvalue + e.last_fvalue) * 0.5f,
-                              d_step == -1, e.stats, c);
-              }
-            }
-          }
-          // update the statistics
-          e.stats.Add(gpair[ridx]);
-          e.last_fvalue = fvalue;
-        }
-      }
-      // finish updating all statistics, check if it is possible to include all sum statistics
-      for (int nid : qexpand) {
-        ThreadEntry &e = temp[nid];
-        c.SetSubstract(snode_[nid].stats, e.stats);
-        if (e.stats.sum_hess >= param_.min_child_weight &&
-            c.sum_hess >= param_.min_child_weight) {
-          bst_float loss_chg;
-          GradStats left_sum;
-          GradStats right_sum;
-          if (d_step == -1) {
-            left_sum = c;
-            right_sum = e.stats;
-          } else {
-            left_sum = e.stats;
-            right_sum = c;
-          }
-          loss_chg = static_cast<bst_float>(
-              spliteval_->ComputeSplitScore(nid, fid, left_sum, right_sum) -
-              snode_[nid].root_gain);
-          const bst_float gap = std::abs(e.last_fvalue) + kRtEps;
-          const bst_float delta = d_step == +1 ? gap: -gap;
-          e.best.Update(loss_chg, fid, e.last_fvalue + delta, d_step == -1, left_sum, right_sum);
-        }
-      }
-    }
-
     // update the solution candidate
     virtual void UpdateSolution(const SparsePage &batch,
-                                const std::vector<int> &feat_set,
+                                const std::vector<bst_feature_t> &feat_set,
                                 const std::vector<GradientPair> &gpair,
                                 DMatrix*p_fmat) {
-      const MetaInfo& info = p_fmat->Info();
       // start enumeration
       const auto num_features = static_cast<bst_omp_uint>(feat_set.size());
 #if defined(_OPENMP)
       const int batch_size =  // NOLINT
           std::max(static_cast<int>(num_features / this->nthread_ / 32), 1);
 #endif  // defined(_OPENMP)
-      int poption = param_.parallel_option;
-      if (poption == 2) {
-        poption = static_cast<int>(num_features) * 2 < this->nthread_ ? 1 : 0;
-      }
-      if (poption == 0) {
-        #pragma omp parallel for schedule(dynamic, batch_size)
+
+      {
+        std::vector<float> densities(num_features);
+        CHECK_EQ(feat_set.size(), num_features);
         for (bst_omp_uint i = 0; i < num_features; ++i) {
-          int fid = feat_set[i];
-          const int tid = omp_get_thread_num();
+          bst_feature_t const fid = feat_set[i];
+          densities.at(i) = p_fmat->GetColDensity(fid);
+        }
+
+#pragma omp parallel for schedule(dynamic, batch_size)
+        for (bst_omp_uint i = 0; i < num_features; ++i) {
+          bst_feature_t const fid = feat_set[i];
+          int32_t const tid = omp_get_thread_num();
           auto c = batch[fid];
           const bool ind = c.size() != 0 && c[0].fvalue == c[c.size() - 1].fvalue;
-          if (param_.NeedForwardSearch(p_fmat->GetColDensity(fid), ind)) {
+          auto const density = densities[i];
+          if (colmaker_train_param_.NeedForwardSearch(
+                  param_.default_direction, density, ind)) {
             this->EnumerateSplit(c.data(), c.data() + c.size(), +1,
-                                 fid, gpair, info, stemp_[tid]);
+                                 fid, gpair, stemp_[tid]);
           }
-          if (param_.NeedBackwardSearch(p_fmat->GetColDensity(fid), ind)) {
+          if (colmaker_train_param_.NeedBackwardSearch(param_.default_direction)) {
             this->EnumerateSplit(c.data() + c.size() - 1, c.data() - 1, -1,
-                                 fid, gpair, info, stemp_[tid]);
+                                 fid, gpair, stemp_[tid]);
           }
         }
-      } else {
-        for (bst_omp_uint fid = 0; fid < num_features; ++fid) {
-          this->ParallelFindSplit(batch[fid], fid,
-                                  p_fmat, gpair);
-        }
       }
     }
     // find splits at current level, do split per level
@@ -632,14 +463,14 @@ class ColMaker: public TreeUpdater {
                           DMatrix *p_fmat,
                           RegTree *p_tree) {
       auto feat_set = column_sampler_.GetFeatureSet(depth);
-      for (const auto &batch : p_fmat->GetSortedColumnBatches()) {
+      for (const auto &batch : p_fmat->GetBatches<SortedCSCPage>()) {
         this->UpdateSolution(batch, feat_set->HostVector(), gpair, p_fmat);
       }
       // after this each thread's stemp will get the best candidates, aggregate results
       this->SyncBestSolution(qexpand);
       // get the best result, we can synchronize the solution
       for (int nid : qexpand) {
-        NodeEntry &e = snode_[nid];
+        NodeEntry const &e = snode_[nid];
         // now we know the solution in snode[nid], set split
         if (e.best.loss_chg > kRtEps) {
           bst_float left_leaf_weight =
@@ -651,7 +482,7 @@ class ColMaker: public TreeUpdater {
           p_tree->ExpandNode(nid, e.best.SplitIndex(), e.best.split_value,
                              e.best.DefaultLeft(), e.weight, left_leaf_weight,
                              right_leaf_weight, e.best.loss_chg,
-                             e.stats.sum_hess);
+                             e.stats.sum_hess, 0);
         } else {
           (*p_tree)[nid].SetLeaf(e.weight * param_.learning_rate);
         }
@@ -669,7 +500,7 @@ class ColMaker: public TreeUpdater {
       // so that they are ignored in future statistics collection
       const auto ndata = static_cast<bst_omp_uint>(p_fmat->Info().num_row_);
 
-      #pragma omp parallel for schedule(static)
+#pragma omp parallel for schedule(static)
       for (bst_omp_uint ridx = 0; ridx < ndata; ++ridx) {
         CHECK_LT(ridx, position_.size())
             << "ridx exceed bound " << "ridx="<<  ridx << " pos=" << position_.size();
@@ -711,11 +542,11 @@ class ColMaker: public TreeUpdater {
       }
       std::sort(fsplits.begin(), fsplits.end());
       fsplits.resize(std::unique(fsplits.begin(), fsplits.end()) - fsplits.begin());
-      for (const auto &batch : p_fmat->GetSortedColumnBatches()) {
+      for (const auto &batch : p_fmat->GetBatches<SortedCSCPage>()) {
         for (auto fid : fsplits) {
           auto col = batch[fid];
           const auto ndata = static_cast<bst_omp_uint>(col.size());
-          #pragma omp parallel for schedule(static)
+#pragma omp parallel for schedule(static)
           for (bst_omp_uint j = 0; j < ndata; ++j) {
             const bst_uint ridx = col[j].index;
             const int nid = this->DecodePosition(ridx);
@@ -748,6 +579,7 @@ class ColMaker: public TreeUpdater {
     }
     //  --data fields--
     const TrainParam& param_;
+    const ColMakerTrainParam& colmaker_train_param_;
     // number of omp thread used during training
     const int nthread_;
     common::ColumnSampler column_sampler_;
@@ -761,26 +593,35 @@ class ColMaker: public TreeUpdater {
     std::vector<int> qexpand_;
     // Evaluates splits and computes optimal weights for a given split
     std::unique_ptr<SplitEvaluator> spliteval_;
+
+    FeatureInteractionConstraintHost interaction_constraints_;
   };
 };
 
 // distributed column maker
 class DistColMaker : public ColMaker {
  public:
-  void Init(const std::vector<std::pair<std::string, std::string> >& args) override {
-    param_.InitAllowUnknown(args);
-    pruner_.reset(TreeUpdater::Create("prune"));
-    pruner_->Init(args);
+  void Configure(const Args& args) override {
+    param_.UpdateAllowUnknown(args);
+    pruner_.reset(TreeUpdater::Create("prune", tparam_));
+    pruner_->Configure(args);
     spliteval_.reset(SplitEvaluator::Create(param_.split_evaluator));
-    spliteval_->Init(args);
+    spliteval_->Init(&param_);
   }
+
+  char const* Name() const override {
+    return "distcol";
+  }
+
   void Update(HostDeviceVector<GradientPair> *gpair,
               DMatrix* dmat,
               const std::vector<RegTree*> &trees) override {
     CHECK_EQ(trees.size(), 1U) << "DistColMaker: only support one tree at a time";
     Builder builder(
       param_,
-      std::unique_ptr<SplitEvaluator>(spliteval_->GetHostClone()));
+      colmaker_param_,
+      std::unique_ptr<SplitEvaluator>(spliteval_->GetHostClone()),
+      interaction_constraints_);
     // build the tree
     builder.Update(gpair->ConstHostVector(), dmat, trees[0]);
     //// prune the tree, note that pruner will sync the tree
@@ -793,8 +634,12 @@ class DistColMaker : public ColMaker {
   class Builder : public ColMaker::Builder {
    public:
     explicit Builder(const TrainParam &param,
-                     std::unique_ptr<SplitEvaluator> spliteval)
-        : ColMaker::Builder(param, std::move(spliteval)) {}
+                     ColMakerTrainParam const& colmaker_train_param,
+                     std::unique_ptr<SplitEvaluator> spliteval,
+                     FeatureInteractionConstraintHost _interaction_constraints)
+        : ColMaker::Builder(param, colmaker_train_param,
+                            std::move(spliteval),
+                            std::move(_interaction_constraints)) {}
     inline void UpdatePosition(DMatrix* p_fmat, const RegTree &tree) {
       const auto ndata = static_cast<bst_omp_uint>(p_fmat->Info().num_row_);
       #pragma omp parallel for schedule(static)
@@ -807,9 +652,6 @@ class DistColMaker : public ColMaker {
         this->position_[ridx] = nid;
       }
     }
-    inline const int* GetLeafPosition() const {
-      return dmlc::BeginPtr(this->position_);
-    }
 
    protected:
     void SetNonDefaultPosition(const std::vector<int> &qexpand, DMatrix *p_fmat,
@@ -836,7 +678,7 @@ class DistColMaker : public ColMaker {
             boolmap_[j] = 0;
         }
       }
-      for (const auto &batch : p_fmat->GetSortedColumnBatches()) {
+      for (const auto &batch : p_fmat->GetBatches<SortedCSCPage>()) {
         for (auto fid : fsplits) {
           auto col = batch[fid];
           const auto ndata = static_cast<bst_omp_uint>(col.size());
@@ -904,6 +746,8 @@ class DistColMaker : public ColMaker {
   TrainParam param_;
   // Cloned for each builder instantiation
   std::unique_ptr<SplitEvaluator> spliteval_;
+
+  FeatureInteractionConstraintHost interaction_constraints_;
 };
 
 XGBOOST_REGISTER_TREE_UPDATER(ColMaker, "grow_colmaker")
